/*
 ******************************
 * Object Oriented Programming in C
 *
 * Author: Laurent Deniau, Laurent.Deniau@cern.ch
 *
 * For more information, please see the paper:
 * http://home.cern.ch/ldeniau/html/oopc/oopc.html
 *
 ******************************
 */

#define IMPLEMENTATION

#include <date.h>

/*
  -------------------------
  date implementation
  -------------------------
*/

typedef enum {
  January, February, March, April, May, June, July,
  August, September, October, November, December } year_months;

static int const mon_len[] =
  { 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

static int const mon_cum_len[] =
  { 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 };

static int
isLeap (int const year)
{
  /* year is considered as a true year (i.e. 2000) */
  return !((year%4) || (!(year%100) && (year%400)));
}

static int
leapYears (int const year1, int const year2)
{
  int y1, y2, ny;

  /* year1 and year2 are considered as true years (i.e. 2000) */

  /* sort years */
  if (year1 <= year2) { y1 = year1; y2 = year2; }
  else                { y1 = year2; y2 = year1; }

  /* initialisation */
  ny = isLeap(y1);

  /* round to previous multiple of 4 */
  y1 -= y1%4;
  y2 -= y2%4;

  /* no leap year between y1 and y2 */
  if (y1==y2) return ny;

  /* increase number of leap years */
  ny += (y2-y1)/4;

  /* round to previous multiple of 100 */
  y1 -= y1%100;
  y2 -= y2%100;

  /* no change of century between y1 and y2 */
  if (y1==y2) return ny;

  /* decrease number of leap years */
  ny -= (y2-y1)/100;

  /* round to previous multiple of 400 */
  y1 -= y1%400;
  y2 -= y2%400;

  /* no century multiple of 400 between y1 and y2 */
  if (y1==y2) return ny;

  /* increase number of leap years */
  ny += (y2-y1)/400;

  return ny;
}

int
constMethodDecl(isValid)
{
  if (!(0<=this->m.date.tm_hour && this->m.date.tm_hour<=23) ||
      !(0<=this->m.date.tm_min  && this->m.date.tm_min <=59) ||
      !(0<=this->m.date.tm_sec  && this->m.date.tm_sec <=59) ||
      !(0<=this->m.date.tm_year) ||
      !(0<=this->m.date.tm_mon  && this->m.date.tm_mon <=11) ||
      !(0<=this->m.date.tm_mday && 
	this->m.date.tm_mday<mon_len[this->m.date.tm_mon]))
    return 0;

  /* case of leap years */
  if (this->m.date.tm_mon  == February &&
      this->m.date.tm_mday == 29 &&
      !isLeap(this->m.date.tm_year+1900))
    return 0;

  return 1;
}

void
methodDecl_(setFormat) char const*const format __
{
  if (format) this->m.format = format;
}

void
methodDecl_(set)
     int const day,  int const month, int const year,
     int const hour, int const min,   int const sec __
{
  this->m.date.tm_mday = day;
  this->m.date.tm_mon  = month-1;
  this->m.date.tm_year = year-1900;
  this->m.date.tm_hour = hour;
  this->m.date.tm_min  = min;
  this->m.date.tm_sec  = sec;
}

void
methodDecl_(setToTm) struct tm const a_date __
{
  this->m.date = a_date;
}

void
methodDecl_(setToDate) t_date const*const a_date __
{
  this->m.date = a_date->m.date;
}

int
methodDecl_(setToString) char const*const a_date, char const*const format __
{
  /* not ISO C but in libc */
  extern char *strptime(const char *s, const char *format, struct tm *tm);

  return !strptime(a_date, format ? format : this->m.format, &this->m.date);
}

void
methodDecl(setToNow)
{
  time_t const t = time(NULL);
  struct tm tm_now = *localtime(&t);

  this->m.date = tm_now;
}

struct tm const
constMethodDecl(getTm)
{
  return this->m.date;
}

char const
constMethodDecl_(*const getString) char const* format __
{
  static char the_date[64]; /* should be enough */
  int n;

  n = strftime(the_date, sizeof(the_date),
	       format ? format : this->m.format, &this->m.date);
  the_date[n] = '\0'; /* safer */

  return the_date;
}

BASEOBJECT_IMPLEMENTATION

  methodName(isValid),
  methodName(setFormat),
  methodName(set),
  methodName(setToTm),
  methodName(setToDate),
  methodName(setToString),
  methodName(setToNow),
  methodName(getTm),
  methodName(getString)

ENDOF_IMPLEMENTATION

initClassDecl() /* required */
{
  objDefault(format) = "%d/%m/%Y %H:%M:%S";
  objDefault(date.tm_year) = 100; /* 01/01/2000 00:00:00 */
}

dtorDecl() {} /* required */

t_date
classMethodDecl_(*const new)
     int const day,  int const month, int const year,
     int const hour, int const min,   int const sec __
{
  t_date *const this = date.alloc();
  sendCMsg_(this, date, set) day, month, year, hour, min, sec __;
  return this;
}

t_date
classMethodDecl_(*const newFromTm) struct tm const a_date __
{
  t_date *const this = date.alloc();
  sendCMsg_(this, date, setToTm) a_date __;
  return this;
}

t_date
classMethodDecl_(*const newFromDate) t_date const*const a_date __
{
  t_date *const this = date.alloc();
  sendCMsg_(this, date, setToDate) a_date __;
  return this;
}

t_date
classMethodDecl_(*const newFromString)
     char const*const a_date, char const*const format __
{
  t_date *this = date.alloc();
  if (!sendCMsg_(this, date, setToString) a_date, format __) {
    free(this);
    this = NULL;
  }
  return this;
}

t_date
classMethodDecl(*const newFromNow)
{
  t_date *this = date.alloc();
  sendCMsg(this, date, setToNow);
  return this;
}

int
classMethodDecl_(compare) t_date const*const date1, t_date const*const date2 __
{
  int res;

  res = date1->m.date.tm_year - date2->m.date.tm_year;
  if (res) return res;

  res = date1->m.date.tm_mon - date2->m.date.tm_mon;
  if (res) return res;

  res = date1->m.date.tm_mday - date2->m.date.tm_mday;
  if (res) return res;

  res = date1->m.date.tm_hour - date2->m.date.tm_hour;
  if (res) return res;

  res = date1->m.date.tm_min - date2->m.date.tm_min;
  if (res) return res;

  return date1->m.date.tm_sec - date2->m.date.tm_sec;
} 

double
classMethodDecl_(diff) t_date const*const date1, t_date const*const date2 __
{
  double ns;
  int n, s, nd=0, one_day=24;
  struct tm d1, d2;

  /* same date */
  n = date.compare(date1, date2);
  if (!n) return 0;

  /* sort date */
  if (n>0) { d1 = date1->m.date; d2 = date2->m.date; s =  1; }
  else     { d1 = date2->m.date; d2 = date1->m.date; s = -1; }

  if (d1.tm_year == d2.tm_year) { /* same year */
    
    if (d1.tm_mon == d2.tm_mon) { /* same month */

      if (d1.tm_mday == d2.tm_mday) /* same day */
        nd = one_day = 0;
      else
        nd = d2.tm_mday-d1.tm_mday-1;

    } else {

      nd += mon_cum_len[d2.tm_mon-1];
      nd -= mon_cum_len[d1.tm_mon];

      /* number of days in months */
      n = mon_len[d1.tm_mon]-d1.tm_mday;
      if (n>0) nd += n;
      nd += d2.tm_mday-1;

      /* adjust leap years */
      if (d1.tm_mon<=February && isLeap(d1.tm_year+1900)) ++nd;
    }

  } else { /* one or more years diff */

    /* number of days between years */
    if (d1.tm_year+1 < d2.tm_year) { 
      int ny = leapYears(d1.tm_year+1900+1, d2.tm_year+1900-1);
      nd = (d2.tm_year-d1.tm_year-1-ny)*365+ny*366;
    }

    /* number of days in years */
    nd += 365-mon_cum_len[d1.tm_mon];
    nd += mon_cum_len[d2.tm_mon-1];
    
    /* number of days in months */
    n = mon_len[d1.tm_mon]-d1.tm_mday;
    if (n>0) nd += n;
    nd += d2.tm_mday-1;

    /* adjust leap years */
    if (d1.tm_mon<=February && isLeap(d1.tm_year+1900)) ++nd;
    if (d2.tm_mon> February && isLeap(d2.tm_year+1900)) ++nd;
  }

  /* compute number of seconds from days */
  ns = 24.0 * 60.0 * 60.0 * nd;

  /* number of seconds in hours */
  ns += (one_day-d1.tm_hour) * 60.0 * 60.0;
  ns -= d1.tm_min  * 60.0;
  ns -= d1.tm_sec;

  ns += d2.tm_hour * 60.0 * 60.0;
  ns += d2.tm_min  * 60.0;
  ns += d2.tm_sec;

  return s*ns;
}

CLASS_IMPLEMENTATION

  methodName(new),
  methodName(newFromTm),
  methodName(newFromDate),
  methodName(newFromString),
  methodName(newFromNow),
  methodName(compare),
  methodName(diff)

ENDOF_IMPLEMENTATION
